package org.acme;

import jakarta.persistence.EntityManager;
import jakarta.persistence.EntityManagerFactory;
import jakarta.persistence.OptimisticLockException;
import org.hibernate.jpa.HibernatePersistenceProvider;
import org.junit.jupiter.api.Test;

import java.util.HashMap;
import java.util.UUID;

import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.fail;

public class VanillaJpaTest {

    @Test
    void test() {
        try(EntityManagerFactory pu = new HibernatePersistenceProvider().createContainerEntityManagerFactory(new Info(), new HashMap<>())) {
            EntityManager em = pu.createEntityManager();
            doInTx(em, () -> em.createNativeQuery("""
                        CREATE TABLE IF NOT EXISTS T_ENTITY (
                            ID                 VARCHAR(36),
                            FIELD              VARCHAR(100)  NOT NULL,
                            VERSION            INT           NOT NULL,
                            PRIMARY KEY(ID)
                        );
                    """).executeUpdate()
            );

            var id = UUID.randomUUID().toString();
            doInTx(em, () -> em.persist(new MyEntity(id, "one", 0)));

            doInTx(em, () -> {
                MyEntity e = em.find(MyEntity.class, id);
                assertEquals("one", e.getField());
                assertEquals(0, e.getVersion());
            });

            // update to v1:

            doInTx(em, () -> {
                MyEntity e = new MyEntity(id, "two", 0);
                em.merge(e);
            });

            doInTx(em, () -> {
                MyEntity e = em.find(MyEntity.class, id);
                assertEquals("two", e.getField());
                assertEquals(1, e.getVersion());
            });

            // merge, but no changes, so no version change is expected

            doInTx(em, () -> {
                MyEntity e = new MyEntity(id, "two", 1); // <<< version in the DB is 1, after the change above
                em.merge(e);
            });

            doInTx(em, () -> {
                MyEntity e = em.find(MyEntity.class, id);
                assertEquals("two", e.getField());
                assertEquals(1, e.getVersion()); // <<< version has NOT changed
            });

            try {
                doInTx(em, () -> {
                    MyEntity e = new MyEntity(id, "three", 0); // <<< an out of date version -> optimistic lock exception
                    em.merge(e);
                });
                fail("no exception");
            }catch (OptimisticLockException e) {
                assertEquals("Row was updated or deleted by another transaction (or unsaved-value mapping was incorrect) : [org.acme.MyEntity#" + id + "]", e.getMessage());
            }

        }
    }

    void doInTx(EntityManager em, Runnable r) {
        try {
            em.getTransaction().begin();
            r.run();
            em.flush();
            em.getTransaction().commit();
        } catch (Exception e) {
            em.getTransaction().rollback();
            throw e;
        } finally {
            em.clear();
        }
    }

    static class MutableObject<T> {
        T val;

        MutableObject(T t) {
            this.val = t;
        }
    }
}
